"""Takes a bunch of Rougail XML dispatched in differents folders
as an input and outputs a Tiramisu's file.

Created by:
EOLE (http://eole.orion.education.fr)
Copyright (C) 2005-2018

Forked by:
Cadoles (http://www.cadoles.com)
Copyright (C) 2019-2021

Silique (https://www.silique.fr)
Copyright (C) 2022

distribued with GPL-2 or later license

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Sample usage::

    >>> from rougail import RougailConvert
    >>> rougail = RougailConvert()
    >>> tiramisu = rougail.save('tiramisu.py')

The Rougail

- loads the XML into an internal RougailObjSpace representation
- visits/annotates the objects
- dumps the object space as Tiramisu string

The visit/annotation stage is a complex step that corresponds to the Rougail
procedures.
"""

from typing import List
from tiramisu import Config

from .i18n import _
from .config import RougailConfig
from .objspace import RougailObjSpace
from .reflector import Reflector
from .tiramisureflector import TiramisuReflector
from .annotator import SpaceAnnotator
from .error import DictConsistencyError
from .providersupplier import provider_supplier
from .utils import normalize_family


class RougailConvert:
    """Rougail object
    """
    def __init__(self,
                 rougailconfig: RougailConfig=None,
                 just_doc: bool=False,
                 ) -> None:
        if rougailconfig is None:
            rougailconfig = RougailConfig
        self.rougailconfig = rougailconfig
        xmlreflector = Reflector(self.rougailconfig)
        self.rougailobjspace = RougailObjSpace(xmlreflector,
                                               self.rougailconfig,
                                               just_doc,
                                               )
        self.internal_functions = self.rougailconfig['internal_functions']
        self.dictionaries = False
        self.annotator = False
        self.reflector = None

    def load_dictionaries(self,
                          path_prefix: str=None,
                          ) -> None:
        self.rougailobjspace.paths.set_path_prefix(normalize_family(path_prefix))
        self._load_dictionaries(self.rougailobjspace.xmlreflector,
                                self.rougailconfig['variable_namespace'],
                                self.rougailconfig['dictionaries_dir'],
                                path_prefix,
                                self.rougailconfig['variable_namespace_description'],
                                )
        for namespace, extra_dir in self.rougailconfig['extra_dictionaries'].items():
            if namespace in ['services', self.rougailconfig['variable_namespace']]:
                msg = _(f'Namespace name "{namespace}" is not allowed')
                raise DictConsistencyError(msg, 21, None)
            self._load_dictionaries(self.rougailobjspace.xmlreflector,
                                    namespace,
                                    extra_dir,
                                    path_prefix,
                                    )
        if hasattr(self.rougailobjspace.space, 'variables'):
            provider_supplier(self.rougailobjspace,
                              path_prefix,
                              )
        self.dictionaries = True

    def _load_dictionaries(self,
                           xmlreflector: Reflector,
                           namespace: str,
                           xmlfolders: List[str],
                           path_prefix: str,
                           namespace_description: str=None,
                           ) -> List[str]:
        for xmlfile, document in xmlreflector.load_dictionaries_from_folders(xmlfolders, self.rougailobjspace.just_doc):
            self.rougailobjspace.xml_parse_document(xmlfile,
                                                    document,
                                                    namespace,
                                                    namespace_description,
                                                    path_prefix,
                                                    )

    def annotate(self):
        if self.annotator:
            raise DictConsistencyError(_('Cannot execute annotate multiple time'), 85, None)
        SpaceAnnotator(self.rougailobjspace)
        self.annotator = True

    def reflexion(self,
                  exclude_imports: list=[],
                  ):
        if not self.dictionaries:
            self.load_dictionaries()
        if not self.annotator:
            self.annotate()
        if self.reflector:
            raise DictConsistencyError(_('Cannot execute reflexion multiple time'), 86, None)
        functions_file = self.rougailconfig['functions_file']
        if not isinstance(functions_file, list):
            functions_file = [functions_file]
        functions_file = [func for func in functions_file if func not in exclude_imports]
        self.reflector = TiramisuReflector(self.rougailobjspace,
                                           functions_file,
                                           self.internal_functions,
                                           self.rougailconfig,
                                           )

    def save(self,
             filename: str,
             ) -> str:
        """Return tiramisu object declaration as a string
        """
        if self.reflector is None:
            self.reflexion()
        output = self.reflector.get_text() + '\n'
        if filename:
            with open(filename, 'w') as tiramisu:
                tiramisu.write(output)
        return output
